--%unit Test.This.Parser
%token '(' ')' number

%right     '='
%left      '+' '-' 
%left      '*' '/' 
%right     dummy
%nonassoc  exp

{
type key_type is (cval, ival, noval); 

type yystype is record
  register : character;
  value    : integer; 
end record;
        
}

%% 

statements : statements statement 
           | 
           ;

statement  : expr end
               { put_line(integer'image($1.value)); }
           | error end 
               { put_line("Try again"); 
                 yyerrok; -- yyclearin;
               }
             statement
           ; 

expr       : expr '+' expr 
               { $$.value :=  $1.value + $3.value; }
           | expr '-' expr 
               { $$.value :=  $1.value - $3.value; }
           | expr '*' expr 
               { $$.value :=  $1.value * $3.value; }
           | expr '/' expr 
               { $$.value :=  $1.value / $3.value; }
           | '-' expr  %prec dummy 
               { $$.value := - $2.value; }
           | '(' expr ')' 
               { $$.value :=  $2.value; }
           | number
               { $$.value :=  value; } 
           ; 

end        : end_of_input 
           ;

%% 

with Test.This.Parser;
procedure Test_Me is
begin
  Test.This.Parser.yyparse;
end;

package Test is end;
package Test.This is end;

package Test.This.Parser is 
    procedure yyparse; 
end; 

with text_io;
use  text_io;
with Test.This.Parser.tokens,
     Test.This.Parser.go_to,
     Test.This.Parser.shift_reduce; 
use  Test.This.Parser.tokens,
     Test.This.Parser.go_to,
     Test.This.Parser.shift_reduce; 

package body Test.This.Parser is 

    Value : constant Integer := 12;

    function to_upper(ch : in character) return character is
    begin
      if ch in 'a'..'z' then
	return character'val( character'pos(ch) - (character'pos('a') -
						   character'pos('A')) );
      else
	return ch;
      end if;
    end to_upper;

    procedure yyerror(s: string) is 
    begin 
        put_line(s); 
    end;  

    function YYLex return Token is
    begin
       return Number;
    end YYLex;
    

##

end;
